from ...config import ENV
from ...types import Message
from .base import BaseMonitor, Tags
from prometheus_client import (
    CollectorRegistry,
    Histogram,
    Counter,
    Summary,
    push_to_gateway,
)


class PrometheusMonitor(BaseMonitor):
    labels = [
        "database",
        "schema",
        "name",
        "resource_type",
        "command",
        "version",
        "env",
    ]

    def __init__(self, host: str, port: str, **kwargs):
        self.host = host
        self.port = port
        self.common_tags = {**kwargs.get("common_tags", {}), **{"env": ENV}}

    def _merge_tags(self, tags: Tags):
        if tags:
            return {**tags, **self.common_tags}

        return self.common_tags

    def initialize(self):
        self.registry = CollectorRegistry()

        self.row_counter = Counter(
            "dbt_rows_moved",
            "Records the number of rows moved",
            self.labels,
            registry=self.registry,
        )

        self.run_time = Summary(
            "dbt_run_time_seconds",
            "Records the time it takes to complete a full dbt run",
            labelnames=self.labels,
            registry=self.registry,
        )
        self.exec_time = Histogram(
            "dbt_execution_time_seconds",
            "Records the time it takes to execute a dbt job",
            labelnames=self.labels,
            registry=self.registry,
        )
        self.freshness = Histogram(
            "dbt_freshness_seconds",
            "Records how fresh a source is",
            labelnames=self.labels,
            registry=self.registry,
        )

    def _get_labels(self, tags: Tags):
        result = []
        if isinstance(tags, dict):
            tags = self._merge_tags(tags)
            for lbl in self.labels:
                result.append(tags.get(lbl, "unknown"))
        return result

    def timed(
        self, name: str, tags: Tags = {}, sample_rate: float = 1, use_ms: bool = None
    ):
        labels = self._get_labels(tags)
        return self.run_time.labels(*labels).time()

    def report_execution_time(self, execution_time: float, tags: Tags):
        labels = self._get_labels(tags)
        return self.exec_time.labels(*labels).observe(execution_time)

    def report_freshness_age(self, age: float, tags: Tags):
        labels = self._get_labels(tags)
        return self.freshness.labels(*labels).observe(age)

    def report_rows_moved(self, rows: int, tags: Tags):
        labels = self._get_labels(tags)
        return self.row_counter.labels(*labels).inc(rows)

    def report(self, msg: Message):
        if msg.reporting and msg.reporting.timing:
            self.report_detailed_timing(msg.reporting.timing, tags=msg.context)

        if msg.reporting and msg.reporting.execution_time:
            self.report_execution_time(msg.reporting.execution_time, tags=msg.context)

        if msg.reporting and msg.reporting.freshness:
            self.report_freshness_age(msg.reporting.freshness.age, tags=msg.context)

        if msg.reporting and msg.reporting.rows:
            self.report_rows_moved(msg.reporting.rows, tags=msg.context)

        return push_to_gateway(f"{self.host}:{self.port}", "dbt", self.registry)



import os
from os.path import dirname, join
from dotenv import load_dotenv

dotenv_path = join(dirname(__file__), ".env")
load_dotenv(dotenv_path, verbose=True)

# Environment
ENV = os.environ.get("ENV", "dev")
SENTRY_DSN = os.environ.get("SENTRY_DSN")
SLACK_URL = os.environ.get("SLACK_URL", None)
APP_VERSION = os.environ.get("IMAGE_VERSION")
NAME = os.environ.get("SERVICE_NAME", "dbt-py")
LOG_LEVEL = os.environ.get("LOG_LEVEL", "debug")
DATADOG_HOST = os.environ.get("DD_HOST", None)
DATADOG_PORT = os.environ.get("DD_STATSD_PORT", None)
SUCCESS_ALERTS = bool(int(os.environ.get("SUCCESS_ALERTS", 0)))
PUSHGATEWAY_HOST = os.environ.get("PUSHGATEWAY_HOST", None)
PUSHGATEWAY_PORT = os.environ.get("PUSHGATEWAY_PORT", None)



import typing as T

from dbt.contracts.results import NodeResult
from dbt.contracts.graph.parsed import (
    ParsedNode,
    ParsedSourceDefinition,
    SeedConfig,
    TestConfig,
    NodeConfig,
    SnapshotConfig,
)
from dbt.contracts.graph.compiled import (
    CompiledDataTestNode,
    CompiledSchemaTestNode,
    CompiledSeedNode,
    CompiledSnapshotNode,
)

from ..logger import GLOBAL_LOGGER as log


def _parse_node(node: ParsedNode) -> T.Dict:
    parsed = dict(
        database=node.database.lower(),
        schema=node.schema.lower(),
        path=node.path.lower(),
        name=node.name.lower(),
        resource_type=node.resource_type.lower(),
        package_name=node.package_name.lower(),
    )

    # only add tags when they exist
    if node.tags:
        parsed["tags"] = ", ".join(node.tags).lower()

    return parsed


def _parse_parsed_node(node: ParsedNode) -> T.Dict:
    parsed = _parse_node(node)
    parsed["filename"] = node.root_path
    parsed["abs_path"] = f"{node.root_path}/{node.build_path}"
    return parsed


def _parse_parsed_source_node(node: ParsedSourceDefinition) -> T.Dict:
    parsed = _parse_node(node)
    parsed["filename"] = node.original_file_path
    parsed["abs_path"] = f"{node.root_path}/{node.original_file_path}"
    return parsed


# TO DO, not important
def _parse_seed_node_config(config: SeedConfig) -> T.Dict:
    return {}


def _parse_test_node_config(config: TestConfig) -> T.Dict:
    return dict(
        materialized=config.materialized.lower(),
    )


def _parse_snapshot_node_config(
    config: SnapshotConfig,
) -> T.Dict:
    node_config = _parse_node_config(config)
    snapshot_config = dict(
        strategy=config.get("stategy", "").lower(),
        unique_key=config.get("unique_key", "").lower(),
        updated_at=config.get('updated_at'),
    )

    return {**node_config, **snapshot_config}


def _parse_node_config(config: NodeConfig) -> T.Dict:
    return dict(
        materialized=config.materialized.lower(),
    )


def parse_node(result: NodeResult) -> T.Dict:
    node = result.node
    config = {}

    log.info(type(node))
    # currently each config function returns the entire config
    # as dict, they are seperated out so we can omit unnecessary
    # values in the future
    if isinstance(node, ParsedSourceDefinition):
        log.info("parsing source definition node")
        parsed_node = _parse_parsed_source_node(node)
        return parsed_node

    parsed_node = _parse_parsed_node(node)
    if isinstance(node, CompiledSeedNode):
        log.debug("parsing seed node config")
        config = _parse_seed_node_config(node.config)
    elif isinstance(node, CompiledDataTestNode) or isinstance(
        node, CompiledSchemaTestNode
    ):
        log.debug("parsing test node config")
        config = _parse_test_node_config(node.config)
    elif isinstance(node, CompiledSnapshotNode):
        log.debug("parsing snapshot node config")
        config = _parse_snapshot_node_config(node.config)
    else:
        log.debug("parsing default node config")
        config = _parse_node_config(node.config)

    return {**parsed_node, **config}



from ..config import SLACK_URL
from ..logger import GLOBAL_LOGGER as log
from .alerting.slack import SlackAlert
from .alerting.dummy import DummyAlert


# Alerting factory
def init():
    if SLACK_URL:
        log.info("Using slack alerting.")
        return SlackAlert(SLACK_URL)
    else:
        log.info("Alerting is not enabled.")
        return DummyAlert()



import logbook
import typing as T
from .base import BaseAlert
from ... import dbt_version
from ...types import Message, Freshness
from ...utils.http import make_external_call
from dbt.contracts.results import TimingInfo
from ...logger import logbook_status_to_color
from ...config import SUCCESS_ALERTS, ENV, NAME
from ...utils.tools import get_execution_datetime


class SlackAlert(BaseAlert):
    def __init__(self, url):
        super(SlackAlert, self).__init__()
        self.url = url
        self.success_alerts = SUCCESS_ALERTS

    @staticmethod
    def _get_issue_priority(level: int) -> str:
        if level == logbook.WARNING:
            return 'warning'

        if level in [logbook.ERROR, logbook.WARNING]:
            return 'high'

        return 'normal'

    @staticmethod
    def _get_message_body(
        title: str,
        message: str,
        error: str,
        attachment: T.Dict,
        timing: T.List[TimingInfo]
    ) -> T.Dict:
        return {
            **attachment, 'fallback': title,
            'title': title,
            'text': message if not error else message + f'\n ```{error}```',
            'ts': get_execution_datetime(timing)
        }

    def _get_freshness_attachment(self, attachment: T.Dict, fresh: Freshness) -> T.Dict:
        attachment['fields'].extend([{
            "title": "Loader",
            "value": fresh.loader,
            "short": True
        }, {
            "title": "Age",
            "value": f"{fresh.age} {fresh.unit}s",
            "short": True
        }])

        return attachment

    def _get_context_attachment(self, attachment: T.Dict, context: T.Dict) -> T.Dict:
        if 'path' in context:
            attachment['fields'].append({
                'title': 'Path',
                'value': context.get('path'),
                'short': True
            })

        if dbt_version:
            attachment = {
                **attachment,
                "footer": f"dbt {dbt_version}",
                "footer_icon": "https://www.getdbt.com/ui/img/graph/avatar.png",
            }

        # Add dbt tags
        if 'tags' in context:
            attachment['fields'].append(
                {"title": "Tags", "value": context.get('tags'), "short": True})

        return attachment

    def alert(self, msg: Message):
        status = logbook_status_to_color(msg.level)
        priority = self._get_issue_priority(msg.level)

        attachment = {
            "color": status,
            "fields": [
                {
                    "title": "Priority",
                    "value": priority.capitalize(),
                    "short": True
                },
                {
                    "title": "Environment",
                    "value": ENV,
                    "short": True
                },
            ],
        }

        if msg.reporting and msg.reporting.freshness:
            attachment = self._get_freshness_attachment(attachment, msg.reporting.freshness)

        attachment = self._get_context_attachment(attachment, msg.context)
        msg_body = self._get_message_body(msg.title, msg.message, msg.error, attachment, msg.reporting.timing)
        payload = {'attachments': [msg_body]}

        # Logic to block success messages if not requested
        if (msg.level < logbook.NOTICE) and not self.success_alerts:
            return

        return make_external_call(
            'POST',
            self.url,
            json=payload,
            timeout=30,
            service_name=NAME,
            raise_exception=True
        )



import os
import json
from mock import MagicMock
from dbt.contracts.results import RunResult, SourceFreshnessResult
from dbt.contracts.graph.compiled import CompiledNode, CompiledTestNode


def _load_file(fn):
    basedir = os.path.dirname(os.path.realpath(__file__))
    return json.load(open(os.path.join(basedir, 'fixtures', fn), 'r'))


def _load_freshness_file(fn):
    obj = _load_file(fn)
    return SourceFreshnessResult.from_dict(obj)


def _load_run_file(fn):
    obj = _load_file(fn)
    return RunResult.from_dict(obj)


def _load_test_file(fn):
    obj = _load_file(fn)
    return RunResult.from_dict(obj)


class ResultMock(MagicMock):
    load_snaphot_fixture = _load_run_file('snapshot_result.json')
    load_freshness_fixture = _load_freshness_file('freshness_result.json')
    load_freshness_fixture_warn = _load_freshness_file('freshness_result_warn.json')
    load_freshness_fixture_fail = _load_freshness_file('freshness_result_fail.json')
    load_model_fixture = _load_run_file('model_result.json')
    load_model_fixture_fail = _load_run_file('model_result_fail.json')
    load_test_fixture = _load_test_file('test_result.json')
    load_test_fixture_fail = _load_test_file('test_result_fail.json')
    load_test_fixture_warn = _load_test_file('test_result_warn.json')



import typing as T
from ...types import Message
from abc import ABC, abstractmethod
from dbt.contracts.results import TimingInfo

Tags = T.Optional[T.Union[T.List, T.Dict]]


class BaseMonitor(ABC):
    common_tags = []

    @abstractmethod
    def initialize(self):
        raise NotImplementedError()

    def increment(self, name: str, value: int = 1, tags: Tags = None, sample_rate: float = 1):
        raise NotImplementedError()

    def decrement(self, name: str, value: int = 1, tags: Tags = None, sample_rate: float = 1):
        pass

    def timed(self, name: str, tags: Tags = None, sample_rate: float = 1, use_ms: bool = None):
        pass

    def timing(self, name: str, value: float, tags: Tags = None, sample_rate: float = 1):
        pass

    def _format_tags(self, tags: Tags):
        # For safety
        if isinstance(tags, dict):
            dd_tags = []
            for key in tags.keys():
                # Only keep top level keys
                if isinstance(tags[key], (dict, list)):
                    continue
                dd_tags.append(f'{key}:{tags[key]}')

            return dd_tags

        if isinstance(tags, list):
            return tags

        if isinstance(tags, str):
            return [tags]

        return []

    def _merge_tags(self, tags: Tags):
        formatted_tags = self._format_tags(tags)
        common_tags = self._format_tags(self.common_tags)
        if common_tags and formatted_tags:
            return list(set(common_tags + formatted_tags))

        if common_tags:
            return common_tags

        if formatted_tags:
            return formatted_tags

        return None

    def report_detailed_timing(self, timing: T.List[TimingInfo], tags: Tags = None):
        if len(timing) < 1:
            return

        for timer in timing:
            execution_time = (timer.completed_at - timer.started_at).total_seconds()
            self.timing(f'dbt.{timer.name}.time',
                        execution_time, tags=self._merge_tags(tags))

    def report_execution_time(self, execution_time: float, tags: Tags = None):
        return self.timing('dbt.run.time', execution_time, tags=self._merge_tags(tags))

    def report_freshness_age(self, age: float, tags: Tags = None):
        return self.timing('dbt.freshness.age', age, tags=self._merge_tags(tags))

    def report_rows_moved(self, rows: int, tags: Tags = None):
        metric_name = 'dbt.rows.moved'
        return self.increment(metric_name, rows, tags=self._merge_tags(tags))

    def report(self, msg: Message):
        raise NotImplementedError()



from functools import wraps

from .base import BaseMonitor, Tags
from ...types import Message


class DummyContextManagerDecorator(object):
    def __init__(self, **kwargs):
        pass

    def __call__(self, func):
        # Others
        @wraps(func)
        def wrapped(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapped


class DummyMonitor(BaseMonitor):
    def initialize(self):
        pass

    def decrement(self, name: str, value: int = 1, tags: Tags = None, sample_rate: float = 1):
        pass

    def report(self, msg: Message):
        pass

    def timed(self, name: str, tags: Tags = None, sample_rate: float = 1, use_ms: bool = None):
        return DummyContextManagerDecorator(name=name, tags=tags, sample_rate=sample_rate, use_ms=use_ms)



import pytest
import logbook

from pydbt import logger
from pydbt.constants import StatusColor

from dbt.contracts.results import NodeStatus


@pytest.mark.parametrize(
    "input,expected",
    [
        (NodeStatus.Success, logbook.INFO),
        (NodeStatus.Error, logbook.ERROR),
        (NodeStatus.Fail, logbook.ERROR),
        (NodeStatus.Warn, logbook.WARNING),
        (NodeStatus.Skipped, logbook.WARNING),
        (NodeStatus.Pass, logbook.INFO),
        (NodeStatus.RuntimeErr, logbook.CRITICAL),
        ("unknown_status", logbook.NOTSET),
    ],
)
def test_dbt_to_log_status(input, expected):
    result = logger.dbt_to_log_status(input)
    assert result == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        (logbook.CRITICAL, StatusColor.CRITICAL),
        (logbook.ERROR, StatusColor.ERROR),
        (logbook.WARNING, StatusColor.WARNING),
        (logbook.NOTICE, StatusColor.NOTICE),
        (logbook.INFO, StatusColor.INFO),
        (logbook.DEBUG, StatusColor.DEBUG),
        (logbook.NOTSET, StatusColor.DEBUG),
        (5000, StatusColor.INFO),
    ],
)
def test_logbook_status_to_color(input, expected):
    result = logger.logbook_status_to_color(input)
    assert result == expected



import time
import typing as T
from numbers import Real
from copy import deepcopy
from dbt.contracts.results import TimingInfo


def rename_keys(d: T.Dict, keys: T.Dict) -> T.Dict:
    dc = deepcopy(d)
    for k in keys.keys():
        if k in dc.keys():
            dc[keys[k]] = dc.pop(k)

    return dc


def omit_keys(d: T.Dict, keys: T.List) -> T.Dict:
    return {k: v for k, v in d.items() if k not in keys}


def freshness_age_to_unit(age: T.Union[Real, float], unit: str) -> float:
    age = float(age)
    if unit == 'hour':
        return round(age / 3600, 0)

    if unit == 'minute':
        return round(age / 60, 1)

    if unit == 'day':
        return round(age / 86400, 0)

    return round(age, 2)


def get_execution_datetime(timing: T.List[TimingInfo]) -> float:
    unix = None
    for obj in timing:
        if obj.name == 'execute' and obj.completed_at is not None:
            unix = time.mktime(obj.completed_at.timetuple())
    return unix


def get_full_db_id(db: str, schema: str, table: str, capitalize: bool = True) -> str:
    db_id = f'{db}.{schema}.{table}'
    if capitalize:
        return db_id.upper()
    return db_id


def get_elapsed_milliseconds_since(time_started: float) -> float:
    return (time.time() - time_started) * 1000



from ..logger import GLOBAL_LOGGER as log
from .monitors.ddatadog import DatadogMonitor
from .monitors.prometheus import PrometheusMonitor
from .monitors.dummy import DummyMonitor
from ..config import DATADOG_HOST, DATADOG_PORT, PUSHGATEWAY_HOST, PUSHGATEWAY_PORT

# Monitor factory


def init(**kwargs):
    if DATADOG_HOST and DATADOG_PORT:
        log.info('Using datadog monitor.')
        monitor = DatadogMonitor(
            host=DATADOG_HOST,
            port=DATADOG_PORT,
            **kwargs
        )
        monitor.initialize()
        return monitor
    elif PUSHGATEWAY_PORT and PUSHGATEWAY_HOST:
        log.info('Using prometheus monitor.')
        monitor = PrometheusMonitor(
            host=PUSHGATEWAY_HOST,
            port=PUSHGATEWAY_PORT,
            **kwargs
        )
        monitor.initialize()
        return monitor
    else:
        log.info('Instrumentation is not enabled.')
        return DummyMonitor()



from dataclasses import asdict
from dbt.node_types import NodeType
from dbt.contracts.results import (
    RunResult,
    SourceFreshnessResult,
    NodeStatus,
    NodeResult,
)

from .node import parse_node
from ..logger import dbt_to_log_status
from ..types import Freshness, Message, Reporting
from ..utils.tools import freshness_age_to_unit, get_full_db_id


class Formatter:
    @classmethod
    def _get_freshness(cls, result: SourceFreshnessResult) -> Freshness:
        fresh = result.node.freshness
        is_warning = result.status == NodeStatus.Warn
        unit = fresh.warn_after.period if is_warning else fresh.error_after.period
        threshold = fresh.warn_after.count if is_warning else fresh.error_after.count

        return Freshness(
            unit=unit,
            threshold=threshold,
            loader=result.node.loader,
            age=freshness_age_to_unit(result.age, unit),
        )

    @classmethod
    def _get_reporting(cls, result: NodeResult) -> Reporting:
        rows, freshness = None, None
        if isinstance(result, SourceFreshnessResult):
            freshness = cls._get_freshness(result)

        if isinstance(result, RunResult):
            rows = result.adapter_response.get("rows_affected", None)

        return Reporting(
            rows=rows,
            freshness=freshness,
            timing=result.timing,
            execution_time=result.execution_time,
        )

    @classmethod
    def _get_source_components(cls, result: SourceFreshnessResult) -> Message:
        reporting = cls._get_reporting(result)
        freshness = reporting.freshness

        is_error = result.status in (NodeStatus.Error or NodeStatus.RuntimeErr)
        is_warning = result.status == NodeStatus.Warn

        db_id = get_full_db_id(
            result.node.database, result.node.schema, result.node.name
        )

        title_id = "out of date" if (is_error or is_warning) else "up to date"
        title = f"Source freshness {title_id}."

        error = None
        message = f"Source `{db_id}` is up to date"
        if is_error or is_warning:
            modifier = "failure" if is_error else "warning"
            message = f"Source `{db_id}` is out of date, {modifier} threshold exceeded"

            error = (
                f"Freshness of {freshness.age} {freshness.unit}(s) exceeded the"
                + f" {modifier} threshold of {freshness.threshold} {freshness.unit}(s)"
            )

        error = result.message if result.message else error

        return Message(
            title=title,
            message=message,
            error=error,
            reporting=reporting,
            context={**parse_node(result), **asdict(freshness)},
            level=dbt_to_log_status(result.status),
        )

    @classmethod
    def _get_model_components(cls, result: RunResult) -> Message:
        reporting = cls._get_reporting(result)

        depends_on = ", ".join(set(result.node.depends_on.nodes)).upper()
        title = f"{result.node.config.materialized.capitalize()} model {result.status}."
        target = get_full_db_id(
            result.node.database, result.node.schema, result.node.name
        )
        message = f"*[{result.status.upper()}]* model `{target}` that depends on `{depends_on}`"

        return Message(
            title=title,
            message=message,
            error=result.message,
            reporting=reporting,
            context=parse_node(result),
            level=dbt_to_log_status(result.status),
        )

    @classmethod
    def _get_test_components(cls, result: RunResult) -> Message:
        reporting = cls._get_reporting(result)

        error = None
        test_name = result.node.name
        test_database = result.node.database
        title = f"[{result.status.upper()}] Test {test_name}"
        message = (
            f"*[{result.status.upper()}]* Test {test_name} in `{test_database.upper()}`"
        )

        if result.status in (NodeStatus.Warn, NodeStatus.Fail, NodeStatus.Error):
            error = f"Got {result.message} results, expected 0"

        return Message(
            title=title,
            message=message,
            error=error,
            reporting=reporting,
            context=parse_node(result),
            level=dbt_to_log_status(result.status),
        )

    @classmethod
    def _get_snapshot_components(cls, result: RunResult) -> Message:
        reporting = cls._get_reporting(result)
        target = get_full_db_id(
            result.node.database, result.node.schema, result.node.name
        )

        is_error = result.status in (NodeStatus.Error, NodeStatus.Skipped)

        title = f'Snapshot {"failed" if is_error else "completed"}'
        message = f"*[{result.status.upper()}]* snapshot with target `{target}`"

        return Message(
            title=title,
            message=message,
            error=result.message,
            reporting=reporting,
            context=parse_node(result),
            level=dbt_to_log_status(result.status),
        )

    @classmethod
    def format(cls, result: NodeResult) -> Message:
        resource_type = result.node.resource_type
        switcher = {
            NodeType.Source: cls._get_source_components,
            NodeType.Model: cls._get_model_components,
            NodeType.Test: cls._get_test_components,
            NodeType.Snapshot: cls._get_snapshot_components,
        }
        return switcher.get(resource_type, cls._get_model_components)(result)



import time
import requests
import typing as T
from ..logger import GLOBAL_LOGGER as log
from .tools import get_elapsed_milliseconds_since


def make_external_call(
        method: str, url: str, *, service_name: str = 'external service',
        raise_exception: bool = True, log_error: bool = True, **kwargs
) -> T.Union[str, None, Exception]:
    """A generic wrapper for handling external calls better-ish."""
    request_data = {}
    if kwargs.get('json'):
        request_data = kwargs['json']
    elif kwargs.get('params'):
        request_data = kwargs['params']

    start_time = time.time()
    try:
        # set default timeout of 30 secs if a value is not passed in
        if 'timeout' not in kwargs:
            kwargs['timeout'] = 30

        response = requests.request(method, url, **kwargs)
        # check that we get a 2xx response from the external service
        response.raise_for_status()

        # create log payload
        log_payload = {
            'total_time': get_elapsed_milliseconds_since(start_time),
            'request_url': response.url.split('/')[2],
            'response_status': response.status_code
        }

        log.info(f'Call to {service_name} was successful', payload=log_payload)

        if 'json' in response.headers.get('Content-Type'):
            return response.json()
        else:
            return response.text

    except Exception as e:
        # log error
        if log_error:
            log_payload = {
                'total_time': get_elapsed_milliseconds_since(start_time),
                'request_data': request_data
            }

            log.error(
                f'{service_name} error',
                exception=e,
                payload=log_payload
            )

        # raise exception
        if raise_exception:
            raise e
        else:
            return None



from ...types import Message
from ...config import ENV
from .base import BaseMonitor, Tags

from datadog import statsd, initialize
from datadog.dogstatsd.context import TimedContextManagerDecorator


class DatadogMonitor(BaseMonitor):
    def __init__(self, host: str, port: str, **kwargs):
        super(DatadogMonitor, self).__init__()

        self.host = host
        self.port = port
        self.use_ms = False
        self.common_tags = {**kwargs.get('common_tags', {}), **{'env': ENV}}

    def initialize(self):
        return initialize(statsd_host=self.host, statsd_port=self.port, hostname_from_config=False)

    def increment(self, name: str, value: int = 1, tags: Tags = None, sample_rate: float = 1):
        return statsd.increment(name, value, tags=self._merge_tags(tags), sample_rate=sample_rate)

    def decrement(self, name: str, value: int = 1, tags: Tags = None, sample_rate: float = 1):
        return statsd.decrement(name, value, tags=self._merge_tags(tags), sample_rate=sample_rate)

    def timed(self, name: str, tags: Tags = None, sample_rate: float = 1, use_ms: bool = None):
        return TimedContextManagerDecorator(
            self, name, tags=self._merge_tags(tags), sample_rate=sample_rate, use_ms=use_ms)

    def timing(self, name: str, value: int, tags: Tags = None, sample_rate: float = 1):
        return statsd.timing(name, value, tags=self._merge_tags(tags), sample_rate=sample_rate)

    def report(self, msg: Message):
        if msg.reporting and msg.reporting.timing:
            self.report_detailed_timing(msg.reporting.timing, tags=msg.context)

        if msg.reporting and msg.reporting.execution_time:
            self.report_execution_time(msg.reporting.execution_time, tags=msg.context)

        if msg.reporting and msg.reporting.freshness:
            self.report_freshness_age(msg.reporting.freshness.age, tags=msg.context)

        if msg.reporting and msg.reporting.rows:
            self.report_rows_moved(msg.reporting.rows, tags=msg.context)

        return



import typing as T
from dataclasses import dataclass
from dbt.contracts.results import TimingInfo

Status = T.Union[None, str, int, bool]


@dataclass
class Freshness:
    unit: str
    age: float
    loader: str
    threshold: float


@dataclass
class Reporting:
    rows: T.Optional[int]
    execution_time: float
    timing: T.Optional[T.List[TimingInfo]]
    freshness: T.Optional[Freshness]


@dataclass
class Message:
    level: int
    title: str
    error: str
    message: str
    context: T.Dict
    reporting: T.Optional[Reporting]



from enum import Enum


class StrEnum(str, Enum):
    def __str__(self) -> str:
        return self.value

    # https://docs.python.org/3.6/library/enum.html#using-automatic-values
    def _generate_next_value_(name, start, count, last_values):
        return name


class StatusColor(StrEnum):
    CRITICAL = "#0F0F0F"
    FATAL = CRITICAL
    ERROR = "#E01E5A"
    WARNING = "#FF7900"
    WARN = WARNING
    NOTICE = "#FFCC00"
    INFO = "#1890ff"
    DEBUG = "#888888"
    SUCCESS = "#2EB67D"



try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup
    from pkgutil import walk_packages

    def find_packages(path='.', prefix=""):
        """Quick & dirty replacement if setuptools is not available."""
        yield prefix
        prefix = prefix + "."
        for _, name, ispkg in walk_packages(path, prefix):
            if ispkg:
                yield name

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except (IOError, ImportError):
    long_description = "Python dbt executor with custom logging and alerting that fits into your stack"

INSTALL_REQUIRES = [
    'dbt-snowflake==0.19.*', 'datadog==0.40.1', 'python-dotenv==0.17.0',
    'pygments>=2.4.0', 'sentry-sdk==1.1.0', "prometheus-client==0.11.0"
]

TEST_REQUIRES = [
    'flake8~=3.9.2',
    'pytest~=6.2.4',
    'pytest-cov~=2.12.1',
    'yapf~=0.21.0',
    'mock~=2.0.0',
]

setup(
    name='pydbt',
    version="0.5.2",
    url='https://github.com/henriblancke/dbt-py',
    license='MIT',
    author="Henri Blancke",
    author_email="blanckehenri@gmail.com",
    description='Python dbt executor with customizable logging and alerting',
    long_description=long_description,
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=INSTALL_REQUIRES,
    test_requires=TEST_REQUIRES,
    extras_require={'dev': TEST_REQUIRES},
    scripts=[
        'scripts/pydbt',
    ],
)



import re
import sys
import logbook
import sentry_sdk
import typing as T
from sentry_sdk.hub import Hub
from logbook import NullHandler
from sentry_sdk.utils import (
    to_string,
    get_lines_from_file,
    event_from_exception,
    current_stacktrace,
    capture_internal_exceptions,
)
from dbt.contracts.results import NodeStatus
from dbt.logger import log_manager as dbt_manager, OutputHandler

from .constants import StatusColor
from .config import NAME, ENV, SENTRY_DSN, APP_VERSION

logger = logbook.Logger("pydbt")

sentry_sdk.init(
    dsn=SENTRY_DSN,
    environment=ENV,
    release=f"{NAME}@{APP_VERSION}",
)


def set_sentry_tags(tags: T.Dict):
    for key, val in tags.items():
        sentry_sdk.set_tag(key, str(val))


def dbt_to_log_status(status: NodeStatus, resource_type: T.Optional[str] = None):
    status_mapper = {
        NodeStatus.Success: logbook.INFO,
        NodeStatus.Error: logbook.ERROR,
        NodeStatus.Fail: logbook.ERROR,
        NodeStatus.Warn: logbook.WARNING,
        NodeStatus.Skipped: logbook.WARNING,
        NodeStatus.Pass: logbook.INFO,
        NodeStatus.RuntimeErr: logbook.CRITICAL,
    }

    return status_mapper.get(status, logbook.NOTSET)


def logbook_status_to_color(status: int):
    status_mapper = {
        logbook.CRITICAL: StatusColor.CRITICAL,
        logbook.ERROR: StatusColor.ERROR,
        logbook.WARNING: StatusColor.WARNING,
        logbook.NOTICE: StatusColor.NOTICE,
        logbook.INFO: StatusColor.INFO,
        logbook.DEBUG: StatusColor.DEBUG,
        logbook.NOTSET: StatusColor.DEBUG,
    }

    return status_mapper.get(status, StatusColor.INFO)


class SentryHandler(logbook.Handler):
    def emit(self, record):
        with capture_internal_exceptions():
            self.format(record)
            return self._emit(record)

    def _emit(self, record):
        hub = Hub.current
        if hub.client is None:
            return

        client_options = hub.client.options
        trace = record.kwargs.pop("trace", None)

        # exc_info might be None or (None, None, None)
        if record.exc_info is not None and record.exc_info[0] is not None:
            event, hint = event_from_exception(
                record.exc_info,
                client_options=client_options,
                mechanism={"type": "logbook", "handled": True},
            )
        elif record.exc_info and record.exc_info[0] is None:
            event = {}
            hint = {}
            with capture_internal_exceptions():
                event["threads"] = {
                    "values": [
                        {
                            "stacktrace": current_stacktrace(
                                client_options["with_locals"]
                            ),
                            "crashed": False,
                            "current": True,
                        }
                    ]
                }
        else:
            event = {}
            hint = {}

        if trace:
            context = record.kwargs.pop("context", {})
            pre_context, context_line, post_context = get_lines_from_file(
                context.get("abs_path", None), 1
            )
            event["exception"] = {
                "type": to_string(record.msg),
                "value": trace,
                "stacktrace": {
                    "frames": [
                        {
                            "function": "run",
                            "module": "dbt",
                            "pre_context": pre_context,
                            "context_line": to_string(trace),
                            "post_context": post_context,
                            "abs_path": context.get("abs_path", None),
                            "filename": context.get("filename", None),
                            "lineno": 1,
                            "in_app": True,
                        }
                    ],
                },
            }

        hint["log_record"] = record

        event["level"] = logbook.get_level_name(record.level).lower()
        event["type"] = logbook.get_level_name(record.level).lower()
        event["logger"] = record.channel
        event["logentry"] = {"message": to_string(record.msg), "params": record.args}
        event["extra"] = {
            "lineno": record.lineno,
            "filename": record.filename,
            "function": record.func_name,
            "process": record.process,
            "process_name": record.process_name,
        }

        hub.capture_event(event, hint=hint)


class FormatMessage(logbook.Processor):
    def process(self, record):
        ansi_escape = re.compile(
            r"""
            \x1B    # ESC
            [@-_]   # 7-bit C1 Fe
            [0-?]*  # Parameter bytes
            [ -/]*  # Intermediate bytes
            [@-~]   # Final byte
        """,
            re.VERBOSE,
        )
        clean = ansi_escape.sub("", str(record.msg))
        record.msg = clean.lower()


class LogManager(logbook.NestedSetup):
    def __init__(self, tags=T.Dict, stdout=sys.stdout, stderr=sys.stderr):
        self.tags = tags
        self.stdout = stdout
        self.stderr = stderr

        set_sentry_tags(tags)

        self._null_handler = NullHandler()
        self._dbt_handler = dbt_manager
        self._output_handler = OutputHandler(stdout, bubble=True)
        self._sentry_handler = SentryHandler(level=logbook.WARNING, bubble=True)
        self._format_processor = FormatMessage()
        self._clean_message_processor = CleanMessage()

        self._dbt_handler._file_handler.disabled = True  # hack to disable file logger

        super().__init__(
            [
                self._null_handler,
                self._output_handler,
                self._sentry_handler,
                self._format_processor,
                self._clean_message_processor,
            ]
        )

    def set_level(self, level: str):
        self._output_handler.level = logbook.lookup_level(level.upper())

    def format_json(self):
        for handler in self.objects:
            if hasattr(handler, "format_json"):
                handler.format_json()


class AppendTags(logbook.Processor):
    def __init__(self, tags: T.Dict):
        self.tags = tags
        set_sentry_tags(tags)
        super().__init__()

    def process(self, record):
        record.extra["tags"] = self.tags


class CleanMessage(logbook.Processor):
    def process(self, record):
        record.msg = str(record.msg)
        record.msg = record.msg.replace("*", "")
        record.msg = record.msg.replace("`", "")


GLOBAL_LOGGER = logger



import sys
import time
import http
import dbt.main
import sentry_sdk
import typing as T
from . import dbt_version
from .types import Message
from .handlers import alert, monitor
from multiprocessing import cpu_count
from .parsers.formatter import Formatter
from dbt.contracts.results import RunExecutionResult
from .logger import GLOBAL_LOGGER as log, LogManager, AppendTags


# Hack to silence TCPServer logs
class QuietHandler(http.server.SimpleHTTPRequestHandler):
    def log_message(self, format, *args):
        pass


dbt.main.dbt.task.serve.SimpleHTTPRequestHandler = QuietHandler
# end hack to silence TCPServer logs


def run(command: T.List, tags: T.Dict):
    start = time.time()
    try:
        # Initialize stats and alerting
        alerting = alert.init()
        stats = monitor.init(common_tags=tags)

        @stats.timed("dbt.command.time", sample_rate=0.5)
        def run_command(cmd: T.List):
            return dbt.main.handle_and_check(cmd)

        res: RunExecutionResult
        res, _ = run_command(command)
        for idx, result in enumerate(res.results):
            msg: Message = Formatter.format(result)
            stats.report(msg)
            alerting.alert(msg)
            with AppendTags({**msg.context, **tags}):
                log.log(
                    msg.level,
                    msg.message,
                    trace=msg.error,
                    context=msg.context,
                )

        return sys.exit(0)

    except Exception as err:
        log.error(err)
        sentry_sdk.capture_exception(err)
        return sys.exit(1)
    finally:
        stats.timing("dbt.command.time", time.time() - start)


def main(args: T.List):
    command = args[1:]
    tags = {
        "app": "dbt",
        "command": command[0],
        "version": dbt_version,
        "number_of_cores": cpu_count(),
    }

    log_manager = LogManager(tags=tags)
    log_manager.format_json()

    with log_manager.applicationbound():
        log.info("Starting dbt run")
        with AppendTags(tags).applicationbound():
            return run(command, tags)
